//
// Dynamic "System Service Descriptor Table" kernel-mode address finder
// Author: j00ru//vx
// Date:   2009-11-02
//
// Compilation note: g++ get_ssdt_address.cpp -lpsapi -lntdll
//                   or simply link with both psapi and ntdll libs in Microsoft Visual C++.
//

#include <cstdio>
#include <cstdlib>
#include <string>
#include <windows.h>
#include <psapi.h>

#define IN
#define OUT
using namespace std;

/* The kernel executable module name.
 * It is used both as the LoadLibraryEx argument and
 * during the kernel module list being searched.
 */
const char KernelName[] = "ntoskrnl.exe";

/* Structure describing a single function we want to get
 * some information about.
 *
 * Note: The programmer is supposed to fill only the first two fields,
 *       the latter ones are going to be calculated and put in their places dynamically.
 *       
 *       You must specify valid FunctionName and SyscallId values of the first
 *       function, at least. It must be granted that the function is an exported symbol
 *       of the kernel module, since it is going to be the point of reference for the rest
 *       of the function list.
 *
 *       Moreover, always remember to take care of the last entry, setting all of its values
 *       to NULLs - this is how the parser knowns where to stop.
 */
struct _SearchedFunction
{
  IN  const CHAR* FunctionName;
  OUT WORD  SyscallId;
  OUT DWORD Address;
  OUT DWORD SsdtAddress;
} SearchedFunctions[] = 
{
  {"NtCreateFile"           }, // kernel exported function
  {"NtOpenFile"             }, 
  {"NtQueryIntervalProfile" },
  {NULL                     }  // empty entry
};


/* GetOriginalImageBase routine
 *
 * Function responsible for retrieving the original PE ImageBase field,
 * given a handle to the already loaded module.
 *
 * Since the provided module is already present in the memory, we can assume
 * all the PE structures are valid.
 */
DWORD GetOriginalImageBase(PVOID Module)
{
  PIMAGE_DOS_HEADER DosHeader = (PIMAGE_DOS_HEADER)Module;
  PIMAGE_NT_HEADERS NtHeaders = (PIMAGE_NT_HEADERS)((DWORD)Module + DosHeader->e_lfanew);
  return NtHeaders->OptionalHeader.ImageBase;
}


/* GetDriverBaseAddr routine
 *
 * Function responsible for retrieving the actual driver Base Address.
 */
DWORD GetDriverBaseAddr(const char* BaseName)
{
  static LPVOID BaseAddresses[4096]; // XXX: let's assume there are at most 4096 active device drivers
  DWORD cbNeeded;

  /* Get a list of all the drivers' Image Base Addresses */
  if(!EnumDeviceDrivers(BaseAddresses,sizeof(BaseAddresses),&cbNeeded)) return 0;

  CHAR FileName[MAX_PATH];

  /* Go thru the entire list */
  for( int i=0;i<(int)(cbNeeded/sizeof(LPVOID));i++ )
  {
    /* For each image base, retrieve the driver's name */
    GetDeviceDriverBaseNameA(BaseAddresses[i],FileName,sizeof(FileName));

    /* In case of the current module being kernel, return its base */
    if(!_stricmp(FileName,BaseName)) return (DWORD)BaseAddresses[i];
  }

  /* Should never get here */
  return 0;
}


int main()
{
  HMODULE hKernel;
  PVOID pFunction;

  puts("*** Microsoft WindowsNT Kernel SSDT address finder by j00ru//vx ***\n");

  printf("[+] Loading the %s image into the process'es address space\n",KernelName);
  hKernel = LoadLibraryEx(KernelName,NULL,DONT_RESOLVE_DLL_REFERENCES);
  if(hKernel==NULL)
  {
    printf("[-] LoadLibraryEx failed. LastError: 0x%.8x\n",(UINT)GetLastError());
    return 1;
  }

  printf("[+] Retrieving the exported %s function address\n",SearchedFunctions[0].FunctionName);
  pFunction = (PVOID)GetProcAddress(hKernel,SearchedFunctions[0].FunctionName);
  if(pFunction==NULL)
  {
    printf("[-] GetProcAddress failed. LastError: 0x%.8x\n",(UINT)GetLastError());
    return 1;
  }

  printf("[+] Obtaining some info about the %s module\n",KernelName);
  MODULEINFO KernelInfo;
  if(!GetModuleInformation(GetCurrentProcess(),hKernel,&KernelInfo,sizeof(KernelInfo)))
  {
    printf("[-] GetModuleInformation failed. LastError: 0x%.8x\n",(UINT)GetLastError());
    return 1;
  }
  printf("[+] ImageBase:   0x%.8x\n",(UINT)KernelInfo.lpBaseOfDll);
  printf("[+] SizeOfImage: 0x%.8x\n",(UINT)KernelInfo.SizeOfImage);
  
  /* Calculate both the 'original' (according to the PE header) and current kernel image address ranges */
  DWORD OrgKernelStart   = GetOriginalImageBase(hKernel), OrgKernelEnd = OrgKernelStart + KernelInfo.SizeOfImage;
  DWORD KernelImageStart = (DWORD)hKernel, KernelImageEnd = KernelImageStart + KernelInfo.SizeOfImage;

  SearchedFunctions[0].Address = (DWORD)pFunction - (DWORD)GetModuleHandle(KernelName) + GetOriginalImageBase(hKernel);
  printf("[+] Searching for the 0x%.8x address within 0x%.8x..0x%.8x\n",
      (UINT)SearchedFunctions[0].Address,
      (UINT)KernelImageStart,
      (UINT)KernelImageEnd);
  
  /* Go through the entire local NTOSKRNL memory in search of the first function's address.
   *
   * When found, we're somewhere inside the SSDT itself.
   * Given the offset at which the address has been found, we can calculate the address of any
   * other syscall handler, having its SyscallId.
   */
  for( PUCHAR i=(PUCHAR)KernelImageStart;i<(PUCHAR)KernelImageEnd-sizeof(DWORD);i++ )
  {
    if(( *(DWORD*)(i+0) == SearchedFunctions[0].Address )                     &&
       ( *(DWORD*)(i-4) >= OrgKernelStart && *(DWORD*)(i-4) <= OrgKernelEnd ) && // Since a heuristic approach is used here, 
       ( *(DWORD*)(i+4) >= OrgKernelStart && *(DWORD*)(i+4) <= OrgKernelEnd ) )  // is is reasonble to check if the adjacent
                                                                                 // DWORD values point somewhere inside the kernel image,
                                                                                 // as well.
    {
      printf("[+] Function pointer found at [0x%.8x]\n",(UINT)i);
      SearchedFunctions[0].SsdtAddress = (DWORD)i;
      break;
    }
  }

  /* Get the SyscallId values for each function from the user-mode (ntdll.dll) code
   */
  for( ULONG i=0;SearchedFunctions[i].FunctionName;i++ )
  {
    HMODULE hNtdll = GetModuleHandle("ntdll.dll");
    FARPROC pFunc  = GetProcAddress(hNtdll,SearchedFunctions[i].FunctionName);
  
    /* Ignore invalid entries
     */
    if(pFunc==NULL)
      continue;

    SearchedFunctions[i].SyscallId = *(DWORD*)(((DWORD)pFunc)+1);
  }

  /* Recalculate the Address values for any syscall function we're interested in
   */
  for( ULONG i=0;SearchedFunctions[i].FunctionName;i++ )
  {
    SearchedFunctions[i].SsdtAddress = SearchedFunctions[0].SsdtAddress + 
                                       (SearchedFunctions[i].SyscallId - SearchedFunctions[0].SyscallId)*sizeof(DWORD);

    SearchedFunctions[i].Address = *(DWORD*)SearchedFunctions[i].SsdtAddress - OrgKernelStart + GetDriverBaseAddr(KernelName);

    printf("[+] < %28s @ 0x%.8x > (SyscallId: 0x%.4x)\n",
        SearchedFunctions[i].FunctionName,
        (UINT)SearchedFunctions[i].Address,
        (USHORT)SearchedFunctions[i].SyscallId);
  }

  return 0;
}

